unit NTPTimeTester;

{
Revision history:

1.0.0  2009 Dec 08  First version
1.0.1  2009 Dec 14  Add composite server TX/RX time plot
1.0.2  2009 Dec 17  Add reference and local NTP version reports

Copyright  2002-2009 by David J Taylor, Edinburgh
Web site:  www.satsignal.eu
E-mail:    davidtaylor@writeme.com
}


interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, IdBaseComponent, IdComponent, IdUDPBase, IdUDPClient, StdCtrls,
  NTPdefinitions, ExtCtrls, TeeProcs, TeEngine, Chart, Series, djtEdit,
  PJVersionInfo;

const
  MaxTests = 2500;

type
  real = double;

type
  TFormNTPtime = class (TForm)
    ButtonClose: TButton;
    MemoDiags: TMemo;
    ButtonGo: TButton;
    IdUDPClientGetNTP: TIdUDPClient;
    Chart1: TChart;
    Series1: TFastLineSeries;
    Chart2: TChart;
    FastLineSeries1: TFastLineSeries;
    LabelServer: TLabel;
    EditServerName: TdjtEdit;
    Chart3: TChart;
    FastLineSeries2: TFastLineSeries;
    Chart4: TChart;
    FastLineSeries3: TFastLineSeries;
    Series2: TFastLineSeries;
    EditReference: TdjtEdit;
    LabelReference: TLabel;
    IdUDPClientRef: TIdUDPClient;
    ButtonClockSeq: TButton;
    LabelResults: TLabel;
    ButtonHelp: TButton;
    LabelRefVersion: TLabel;
    LabelTestVersion: TLabel;
    PJVersionInfo: TPJVersionInfo;
    procedure ButtonClockSeqClick (Sender: TObject);
    procedure ButtonCloseClick (Sender: TObject);
    procedure ButtonGoClick (Sender: TObject);
    procedure ButtonHelpClick (Sender: TObject);
    procedure FormCreate (Sender: TObject);
    procedure FormResize (Sender: TObject);
  private
    { Private declarations }
    FBytesRefIn: TBytes;
    FBytesRefOut: TBytes;
    FBytesTestIn: TBytes;
    FBytesTestOut: TBytes;
    FClientTX, FClientRX: array [0..MaxTests-1] of double;
    FNumTests: integer;
    FRefTimes: array [0..MaxTests-1] of double;
    FServerRX, FServerTX: array [0..MaxTests-1] of double;
    FSNTPRefPacket: TSNTPpacket;
    FSNTPTestPacket: TSNTPpacket;
    FStartTime: TDateTime;
    FStarted: boolean;
    procedure AnalyseData;
    procedure CheckGetNTP;
    procedure GetNTPversion (udp: TIdUDPClient;  var version: string);
  public
    { Public declarations }
  end;

var
  FormNTPtime: TFormNTPtime;


implementation

{$R *.dfm}

uses
  MMSystem, Math, CurveFit, Utils;


procedure TFormNTPtime.AnalyseData;
const
  MinCells = -120;
  MaxCells = +200;
var
  index: integer;
  i_delay, i_round: integer;
  server_min_delay, server_max_delay, delay: double;
  round_min_delay, round_max_delay, round_trip: double;
  max_delay, max_round: double;
  histo_delay, histo_round: array [MinCells..MaxCells] of integer;
  server_delay: array [0..MaxTests-1] of double;
  round_delay: array [0..MaxTests-1] of double;
  server_tx, server_rx: array [0..MaxTests-1] of real;
  expected: double;
begin
  server_min_delay := 1E30;
  server_max_delay := -server_min_delay;
  round_min_delay := server_min_delay;
  round_max_delay := server_max_delay;

  for index := 0 to MaxTests - 1 do
    begin
    server_delay [index] := FServerTX [index] - FServerRX [index];
    server_min_delay := Min (server_min_delay, server_delay [index]);
    server_max_delay := Max (server_max_delay, server_delay [index]);
    round_delay [index] := FClientRX [index] - FClientTX [index];
    round_min_delay := Min (round_min_delay, round_delay [index]);
    round_max_delay := Max (round_max_delay, round_delay [index]);
    end;

  if round_max_delay < 2000 then max_round := 2000 else
  if round_max_delay < 5000 then max_round := 5000 else
  if round_max_delay < 10000 then max_round := 10000 else
  if round_max_delay < 20000 then max_round := 20000 else max_round := 50000;

  if server_max_delay < 5 then max_delay := 5 else
  if server_max_delay < 10 then max_delay := 10 else
  if server_max_delay < 20 then max_delay := 20 else
  if server_max_delay < 50 then max_delay := 50 else
  if server_max_delay < 100 then max_delay := 100 else
  if server_max_delay < 200 then max_delay := 200 else
  if server_max_delay < 500 then max_delay := 500 else
  if server_max_delay < 1000 then max_delay := 1000 else
  if server_max_delay < 2000 then max_delay := 2000 else max_delay := 5000;

  for index := MinCells to MaxCells do
    begin
    histo_delay [index] := 0;
    histo_round [index] := 0;
    end;

  for index := 0 to MaxTests - 1 do
    begin
    i_round := Round (round_delay [index] * MaxCells / max_round);
    if i_round < MinCells then i_round := MinCells;
    if i_round > MaxCells then i_round := MaxCells;
    Inc (histo_round [i_round]);

    i_delay := Round (server_delay [index] * MaxCells / max_delay);
    if i_delay < MinCells then i_delay := MinCells;
    if i_delay > MaxCells then i_delay := MaxCells;
    Inc (histo_delay [i_delay]);
    end;
  Chart1.Series [0].Clear;
  Chart3.Series [0].Clear;
  for index := MinCells to MaxCells do
    begin
    delay := index * max_delay / MaxCells;
    Chart1.Series [0].AddXY (delay, histo_delay [index]);
    Chart3.Series [0].AddXY (delay, histo_delay [index]);
    end;
  Chart2.Series [0].Clear;
  for index := MinCells to MaxCells do
    begin
    round_trip := index * max_round / MaxCells;
    Chart2.Series [0].AddXY (round_trip, histo_round [index]);
    end;
  Chart4.Series [0].Clear;
  Chart4.Series [1].Clear;
  for index := 0 to MaxTests - 1 do
    begin
    server_tx [index] := FServerTX [index] - FStartTime;
    server_rx [index] := FServerRX [index] - FStartTime;
    end;
  for index := 0 to MaxTests - 1 do
    begin
    expected := FRefTimes [index];
    Chart4.Series [0].Add (FServerRX [index] - expected);
    Chart4.Series [1].Add (FServerTX [index] - expected);
    end;
end;


procedure TFormNTPtime.ButtonClockSeqClick (Sender: TObject);
type
  ftime = packed record
    case integer of
      1: (ft: TFileTime);
      2: (i64: int64)
  end;
var
  f0, f1, f2: ftime;
  count: integer;
  t: TDateTime;
  fail: boolean;
begin
  MemoDiags.Clear;
  MemoDiags.Update;
  MemoDiags.Lines.Add ('Clock sequence test');

  count := 0;
  GetSystemTimeAsFileTime (f0.ft);
  f1 := f0;
  t := f1.i64;
  t := t / (1E7 * 86400);
  t := t + EncodeDate (1601, 1, 1);
  MemoDiags.Lines.Add (FormatDateTime ('"  start st " yyyy-mm-dd hh:nn:ss', t));
  repeat
    GetSystemTimeAsFileTime (f2.ft);
    fail := f2.i64 < f1.i64;
    if fail
    then
      begin
      MemoDiags.Lines.Add ('Clock sequence test failed');
      MemoDiags.Lines.Add ('Clock sequence:');
      MemoDiags.Lines.Add (Format ('    %d iterations', [count]));
      MemoDiags.Lines.Add (Format ('    %.4f ms', [(f2.i64 - f1.i64)/1E4]));
      end
    else
      begin
      Inc (count);
      f1 := f2;
      end;
  until (count > 100000000) or fail;
  if not fail then
    begin
    MemoDiags.Lines.Add ('Clock sequence test passed');
    MemoDiags.Lines.Add (Format ('Clock sequence: %d iterations', [count]));
    end;
end;


procedure TFormNTPtime.ButtonCloseClick (Sender: TObject);
begin
  Close;
end;


procedure TFormNTPtime.ButtonGoClick (Sender: TObject);
begin
  MemoDiags.Clear;
  FStartTime := NowUTC;
  FStarted := False;
  FNumTests := 0;
  repeat
    CheckGetNTP;
    Inc (FNumTests);
  until FNumTests >= MaxTests;

  AnalyseData;
end;


procedure TFormNTPtime.ButtonHelpClick (Sender: TObject);
var
  help: string;
begin
  help := 'This program has two functions, to compare' + #13 +
          'the RX and TX times in NTP packets returned' + #13 +
          'from an NTP server, and to check whether ' + #13 +
          'sucessive calls to GetSystemTimeAsFileTime' + #13 +
          'ever return backwards-going values.' + #13 + #13 +
          'The Reference Server (top entry) is used to' + #13 +
          'plot the TX and RX times in the returned NTP' + #13 +
          'packets against a local reference, so that' + #13 +
          'one can judge whether the jitter is either' + #13 +
          'in the TX or RX stimestamps.' + #13 +
          '2500 calls are made to the remote NTP server.' + #13 +
          'Press the Go button to run this test.' + #13 + #13 +
          'The clock sequence test runs 100000000 calls' + #13 +
          'and reports any negative time jumps.  Press' + #13 +
          'the Clock Sequence button to run.';
  ShowMessage (help);
end;


procedure TFormNTPtime.CheckGetNTP;
var
  orig, my_rx: TNtpTimestamp;
  client_tx, server_rx, server_tx, client_rx: TDateTime;
  version: string;
begin
  IdUDPClientRef.Host := EditReference.Text;
  if not IdUDPClientRef.Connected then
    begin
    IdUDPClientRef.Connect;
    GetNTPversion (IdUDPClientRef, version);
    LabelRefVersion.Caption := version;
    LabelRefVersion.Hint := version;
    LabelRefVersion.Update;
    end;

  IdUDPClientGetNTP.Host := EditServerName.Text;
  if not IdUDPClientGetNTP.Connected then
    begin
    IdUDPClientGetNTP.Connect;
    GetNTPversion (IdUDPClientGetNTP, version);
    LabelTestVersion.Caption := version;
    LabelTestVersion.Hint := version;
    LabelTestVersion.Update;
    end;

  FillChar (FSNTPRefPacket, SizeOf (FSNTPRefPacket), 0);
  // LeapInd=0  VersionNum=1  Mode=3
  FSNTPRefPacket.flags := (0 shl 6) + (1 shl 3) + 3;

  client_tx := NowUTC;
  orig := DateTimeToNtpTimestamp (client_tx);
  FSNTPRefPacket.transmit_timestamp := orig;   // use the third slot

  System.Move (FSNTPRefPacket, FBytesRefOut [0], SizeOf (FSNTPRefPacket));
  System.Move (FSNTPRefPacket, FBytesTestOut [0], SizeOf (FSNTPTestPacket));
  IdUDPClientRef.SendBuffer (FBytesRefOut);
  IdUDPClientRef.ReceiveBuffer (FBytesRefIn, SizeOf (FSNTPRefPacket));
  client_rx := NowUTC;
  IdUDPClientGetNTP.SendBuffer (FBytesTestOut);
  IdUDPClientGetNTP.ReceiveBuffer (FBytesTestIn, SizeOf (FSNTPTestPacket));
  System.Move (FBytesRefIn [0], FSNTPRefPacket, SizeOf (FSNTPRefPacket));
  System.Move (FBytesTestIn [0], FSNTPTestPacket, SizeOf (FSNTPTestPacket));
  my_rx := DateTimeToNtpTimestamp (client_rx);

  if not FStarted then
    begin
    FStarted := True;
    FNumTests := 0;
    Exit;
    end;

  server_rx := NtpTimestampToDateTime (FSNTPRefPacket.receive_timestamp);
  server_tx := NtpTimestampToDateTime (FSNTPRefPacket.transmit_timestamp);
  FRefTimes [FNumTests] := 0.5 * 86400000000 * (server_rx + server_tx);

  server_rx := NtpTimestampToDateTime (FSNTPTestPacket.receive_timestamp);
  server_tx := NtpTimestampToDateTime (FSNTPTestPacket.transmit_timestamp);

  FClientTX [FNumTests] := 86400000000 * client_tx;
  FServerRX [FNumTests] := 86400000000 * server_rx;
  FServerTX [FNumTests] := 86400000000 * server_tx;
  FClientRX [FNumTests] := 86400000000 * client_rx;
  if FNumTests > 1 then Exit;

  MemoDiags.Clear;
  MemoDiags.Lines.Add ('Ref: ' + NtpTimestampToStr (FSNTPTestPacket.reference_timestamp, True));
  MemoDiags.Lines.Add ('Org: ' + NtpTimestampToStr (FSNTPTestPacket.originate_timestamp, True));
  MemoDiags.Lines.Add (' RX: ' + NtpTimestampToStr (FSNTPTestPacket.receive_timestamp, True));
  MemoDiags.Lines.Add (' TX: ' + NtpTimestampToStr (FSNTPTestPacket.transmit_timestamp, True));
  MemoDiags.Lines.Add ('Rcv: ' + NtpTimestampToStr (my_rx, True));
end;


procedure TFormNTPtime.FormCreate (Sender: TObject);
begin
  DesktopFont := True;
  SetLength (FBytesRefOut, SizeOf (FSNTPRefPacket));
  SetLength (FBytesRefIn, SizeOf (FSNTPRefPacket));
  SetLength (FBytesTestOut, SizeOf (FSNTPTestPacket));
  SetLength (FBytesTestIn, SizeOf (FSNTPTestPacket));
  Application.Title := Caption;
  Caption := Caption + ' - V' +  PJVersionInfo.FileVersion;
end;


procedure TFormNTPtime.FormResize (Sender: TObject);
var
  L: integer;
  dw: integer;
begin
  L := MemoDiags.Left + MemoDiags.Width + 8;
  Chart1.Left := L;
  dw := (ClientWidth - L - 16) div 2;
  Chart1.Width := dw;
  Chart2.Left := Chart1.Left + Chart1.Width + 8;
  Chart2.Width := dw;
  Chart3.Left := Chart1.Left;
  Chart3.Width := Chart1.Width;
  Chart4.Left := Chart2.Left;
  Chart4.Width := Chart2.Width;
end;


procedure TFormNTPtime.GetNTPversion (udp: TIdUDPClient; var version: string);
const
  buf_size = 512;
var
  request: TNTPControlPacket;
  out_pkt, in_pkt: TBytes;
  buf: array [0..buf_size-1] of AnsiChar;
  bytes_rx: integer;
  version_size: integer;
  index: integer;
  ch: AnsiChar;
begin
  version := '';
  FillChar (request, SizeOf (TNTPControlPacket), 0);
  // LeapInd=0  VersionNum=2  Mode=6
  request.flags := (0 shl 6) + (2 shl 3) + 6;
  request.flags2 := READVAR;
  request.flags3 := 1;
  SetLength (out_pkt, SizeOf (TNTPControlPacket));
  System.Move (request, out_pkt [0], SizeOf (TNTPControlPacket));
  udp.SendBuffer (out_pkt);
  SetLength (in_pkt, buf_size);
  bytes_rx := udp.ReceiveBuffer (in_pkt, buf_size);
  if bytes_rx > 0 then
    begin
    System.Move (in_pkt [0], buf, bytes_rx);
    index := 21;
    repeat
      ch := buf [index];
      if ch <> '"' then version := version + ch;
      Inc (index);
    until (ch = '"') or (Length (version) > 80);
    end;
end;

end.
